from cx_Freeze import setup, Executable

setup(
    name="SBML2HYB",
    version="1.0.1",
    description="DESCRIPTION",
    options={
        "sbml2hyb": {"packages": ["Pillow", "tensorflow", "tkinter", "python-libsbml"]}
    },
    executables=[Executable("sbml2hyb.py", base="Win32GUI")],
)



from tensorflow import keras


def main(parameters, species, species_checkboxes, parameters_checkboxes, h5_file, curr_file_name, hmod_file_string):
    inputs = [] #input IDs for the model network test ["S"]
    for i in range(len(species_checkboxes)):
        if species_checkboxes[i][1].get() == 1:
            inputs.append(species[i])

    hybout = [] #output IDs for the model network test ["miu","vPM","vPT"]
    for i in range(len(parameters_checkboxes)):
        if parameters_checkboxes[i][1].get() == 1:
            hybout.append(parameters[i])

    # open Keras file
    model = keras.models.load_model(h5_file)

    weights = model.get_weights() # list of numpy arrays containing weights in a per layer/bias basis

    # This section orders the weights in a single list for future usage in the format:
    # Input->Hidden1 weights;
    # Input->Hidden1 bias;
    # Hidden1->Hidden2 weights;
    # Hidden1->Hidden2 bias; etc.
    countgroup=0
    countsubgroup=0
    orderedweights=[]
    for groups in weights:
        if (countgroup % 2) == 0:
            for subgroups in weights[countgroup]:
                orderedweights.extend(weights[countgroup][countsubgroup])
                countsubgroup=countsubgroup+1
        else:
            orderedweights.extend(weights[countgroup])
        countsubgroup=0
        countgroup=countgroup+1

    nlayers=len(model.layers) #Check number of model layers
    nin = len(inputs) #Check number of inputs
    countin=0 #Create a parsed input counter
    equations =[] #Create empty equations list to be built
    H=[] #Create empty hidden nodes list
    wcount=0 #number of parsed weights (required for some adjustments)

    for layer in range(nlayers):
        H.append([]) #Add a new layer to list
        nH = len(weights[layer*2+1]) #number of nodes in the layer. Read from the number of bias values, since it is currently assumed to be one per node (Standard dense layers)
        if layer == 0:
            for node in range(nH):
                H[layer].append("")
                #First layer and node is w1*in1 +(w1+nH)*in2 + (w1+nH*2)*in2... w1+nh*nin. Second node sums 1 to each. Third sums 2 to the first and so on.
                for n in range(nin):
                    H[layer][node] = H[layer][node] + ("w" if n==0 else "+w") + str(1+node+nH*n) + "*" + inputs[n]

                    if n==nin-1:
                        H[layer][node] = H[layer][node] + "+w" + str(1+node+nH*(n+1))
                        H[layer][node] = "tanh(" + H[layer][node] + ")"

            wcount=wcount+ len(H[layer])*len(inputs)+len(H[layer]) #add number of parsed on this step nodes to count

        elif 0 < layer < nlayers-1:
            for node in range(nH):
                H[layer].append("")
                #Second to second last layer are based on the previous number of nodes (size of the H in the previous row list) to the next number of nodes(nH)
                for n in range(len(H[layer-1])):
                    H[layer][node] = H[layer][node] + ("w" if n==0 else "+w") + str(wcount+1+node+nH*n) + "*" + H[layer-1][n]

                    if n==len(H[layer-1])-1:
                        H[layer][node] = H[layer][node] + "+w" + str(wcount+1+node+nH*(n+1))
                        H[layer][node] = "tanh(" + H[layer][node] + ")"

            wcount=wcount+ len(H[layer])*len(H[layer-1])+len(H[layer]) #add number of parsed on this step nodes to count

        elif layer == nlayers-1:
            for node in range(nH):
                H[layer].append("")
                #Last layer is based on the previous number of nodes (size of the H in the previous row list) to the number of outputs
                for n in range(len(H[layer-1])):
                    H[layer][node] = H[layer][node] + ("w" if n==0 else "+w") + str(wcount+1+node+nH*n) + "*" + H[layer-1][n]

                    if n==len(H[layer-1])-1:
                        H[layer][node] = H[layer][node] + "+w" + str(wcount+1+node+nH*(n+1))


            wcount=wcount+ len(H[layer])*len(H[layer-1])+len(H[layer]) #add number of parsed on this step nodes to count

    equations = H[-1]
    # Original file and output file
    f = hmod_file_string.split("\n")
    h = ""

    o_name = curr_file_name.split(".hmod")[0] if ".hmod" in curr_file_name else curr_file_name.split(".xml")[0]
    n_name = o_name

    # Prepare automatic variables
    Asscount=0 #counter for assignments
    wcount=0 #counter for weights
    subAss=0 # counter for repeated assignments
    for line in f:
        if ".nparameters=" in line:
            pos=line.index("=")
            nbasepar=int(line[pos+1:-1])#obtain number of base parameters
            nhybpar=int(line[pos+1:-1]) #Define a counter for the hybrid parameters

        elif ".nruleAss" in line:
            pos=line.index("=")
            nbaseAss=int(line[pos+1:-1]) #obtain number of base assignments
            nhybAss=int(line[pos+1:-1]) #Define a counter for the hybrid assignments

        elif (".ruleAss(" in line) and (").id" in line): #check for assignment rules to substitute
            if any(item in line for item in hybout):
                subAss=subAss+1

    skip = 0 # Placeholder for skipped lines
    outhyb = 0 # Placeholder of hybrid output
    #Write hybrid file
    for line in f:
        if ".parameters("+str(nbasepar)+").reaction" in line: #Check if the last standard parameter line has been reached
            if outhyb==1:
                outhyb=0
                h += line.replace(o_name, n_name).replace("local","global") +"\n"
            else:
                h += line.replace(o_name,n_name)+"\n" #write the last standard line for the hybrid model
            for w in orderedweights: #Add all hybrid weights
                nhybpar=nhybpar+1 #Increase hybrid parameter counter. Following 4 lines write the new hybrid parameter
                h += n_name+'.parameters('+str(nhybpar)+').id="w'+str(wcount+1)+'";\n'
                h += n_name+".parameters("+str(nhybpar)+").val="+str(orderedweights[wcount])+";\n"
                h += n_name+".parameters("+str(nhybpar)+").fixed=0;\n"
                h += n_name+'.parameters('+str(nhybpar)+').reaction="global";\n'
                wcount=wcount+1 #Increase weights counter

        elif (".parameters(" in line) and (").id" in line):
            if any(item in line for item in hybout):
                h += line.replace(o_name,n_name)+"\n"
                outhyb=1 #reduced number of new hybrid assignments
            else:
                h += line.replace(o_name,n_name)+"\n"

        elif (".parameters(" in line) and (").reaction" in line):
            if outhyb==1:
                outhyb=0
                h += line.replace(o_name, n_name).replace("local","global")+"\n"
            else:
                h += line.replace(o_name, n_name)+"\n"

        elif ".ruleAss("+str(nbaseAss)+").val" in line or ("nruleAss=0" in line): #Check if the last standard assignment line has been reached
            if skip>0:
                skip=skip-1
            elif skip==0 and ("nruleAss=0" in line):
                h += line.replace(str(nbaseAss),str(nbaseAss+len(hybout)-subAss)).replace(o_name,n_name)+"\n"
            else:
                h += line.replace(o_name,n_name)+"\n" #write the last standard line for the hybrid model
            for par in hybout: #Add all hybrid assignments
                if Asscount >= len(equations):
                    return [None, len(equations)]
                nhybAss=nhybAss+1 #Increase hybrid assignments counter. Following 4 lines write the new hybrid assignments
                h += n_name+'.ruleAss('+str(nhybAss)+').id="'+hybout[Asscount]+'";\n'
                h += n_name+'.ruleAss('+str(nhybAss)+').val="'+equations[Asscount]+'";\n'
                Asscount=Asscount+1 #Increase assignments counter

        elif ".nparameters=" in line: #Correct number of parameters
            h += line.replace(str(nbasepar),str(nbasepar+len(orderedweights))).replace(o_name,n_name)+"\n"

        elif ".nruleAss" in line: #Correct number of assignments
            h += line.replace(str(nbaseAss),str(nbaseAss+len(hybout)-subAss)).replace(o_name,n_name)+"\n"

        elif (".ruleAss(" in line) and (").id" in line): #check for assignment rules to substitute
            if any(item in line for item in hybout):
                nhybAss=nhybAss-1 #reduced number of new hybrid assignments
                skip=1
            else:
                h += line.replace(o_name, n_name)+"\n"

        elif skip >0: #check if line should be skipped
            skip=skip-1

        else: #For all other lines, copy them to the new file
            if "% Events" in line:
                h += "\n"
            h += line.replace(o_name,n_name)+"\n"

    return [h, None]



from libsbml import SBMLValidator as validator
from libsbml import SBMLReader
from libsbml import SBMLDocument


def validate_sbml(file_string):
    doc = SBMLReader.readSBMLFromString(SBMLReader(), file_string)
    val = validator()
    val.setDocument(doc)
    a = val.validate()
    failures = val.getNumFailures()

    if failures > 0:
        error_message = ""
        for i in range(failures):
            error_message += val.getFailure(i).getMessage() + "\n"
        return error_message

    return a



class Specie:
    def __init__(
        self, specie_id, compartment, val, fixed, specie_min, specie_max, isres
    ):
        self.id = specie_id
        self.compartment = compartment
        self.val = val
        self.fixed = fixed
        self.min = specie_min
        self.max = specie_max
        self.isres = isres

    def __str__(self):
        return (
            f"ID: {self.id}\n"
            f"Compartment: {self.compartment}\n"
            f"Val: {self.val}\n"
            f"Fixed: {self.fixed}\n"
            f"Min: {self.min}\n"
            f"Max: {self.max}\n"
            f"Isres: {self.isres}\n"
        )


class Compartment:
    def __init__(
        self, compartment_id, val, fixed, compartment_min, compartment_max, isres
    ):
        self.id = compartment_id
        self.val = val
        self.fixed = fixed
        self.min = compartment_min
        self.max = compartment_max
        self.isres = isres

    def __str__(self):
        return (
            f"ID: {self.id}\n"
            f"Val: {self.val}\n"
            f"Fixed: {self.fixed}\n"
            f"Min: {self.min}\n"
            f"Max: {self.max}\n"
            f"Isres: {self.isres}\n"
        )


class Parameter:
    def __init__(self, parameter_id, reaction, val, fixed):
        self.id = parameter_id
        self.reaction = reaction
        self.val = val
        self.fixed = fixed

    def __str__(self):
        return (
            f"ID: {self.id}\n"
            f"Reaction: {self.reaction}\n"
            f"Val: {self.val}\n"
            f"Fixed: {self.fixed}\n"
        )


class AssignmentRule:
    def __init__(self, assignment_rule_id, val):
        self.id = assignment_rule_id
        self.val = val

    def __str__(self):
        return f"ID: {self.id}\n" f"Val: {self.val}\n"


class Event:
    def __init__(self, event_id, trigger, delay, var1, var2):
        self.id = event_id
        self.trigger = trigger
        self.delay = delay
        self.var1 = var1
        self.var2 = var2

    def __str__(self):
        return (
            f"ID: {self.id}\n"
            f"Trigger: {self.trigger}\n"
            f"Delay: {self.delay}\n"
            f"Var1: {self.var1}\n"
            f"Var2: {self.var2}\n"
        )


class Reaction:
    def __init__(self, reaction_id, rate, values):
        self.id = reaction_id
        self.rate = rate
        self.values = values

    def __str__(self):
        return f"ID: {self.id}\n" f"Rate: {self.rate}\n" f"Values: {self.values}\n"


class RateRule:
    def __init__(self, rate_rule_id, val, rate_rule_min, rate_rule_max, isres):
        self.id = rate_rule_id
        self.val = val
        self.min = rate_rule_min
        self.max = rate_rule_max
        self.isres = isres

    def __str__(self):
        return (
            f"ID: {self.id}\n"
            f"Val: {self.val}\n"
            f"Min: {self.min}\n"
            f"Max: {self.max}\n"
            f"Isres: {self.isres}\n"
        )


class Hmod:
    species = []
    compartments = []
    parameters = []
    assignment_rules = []
    events = []
    reactions = []
    rate_rules = []

    def __init__(self):
        self.species = []
        self.compartments = []
        self.parameters = []
        self.assignment_rules = []
        self.events = []
        self.reactions = []
        self.rate_rules = []

    def add_specie(
        self, specie_id, compartment, val, fixed, specie_min, specie_max, isres
    ):
        self.species.append(
            Specie(specie_id, compartment, val, fixed, specie_min, specie_max, isres)
        )

    def add_compartment(
        self, compartment_id, val, fixed, compartment_min, compartment_max, isres
    ):
        self.compartments.append(
            Compartment(
                compartment_id, val, fixed, compartment_min, compartment_max, isres
            )
        )

    def add_parameter(self, parameter_id, reaction, val, fixed):
        self.parameters.append(Parameter(parameter_id, reaction, val, fixed))

    def add_assignment_rule(self, assignment_rule_id, val):
        self.assignment_rules.append(AssignmentRule(assignment_rule_id, val))

    def add_event(self, event_id, trigger, delay, var1, var2):
        self.events.append(Event(event_id, trigger, delay, var1, var2))

    def add_reaction(self, reaction_id, rate, values):
        self.reactions.append(Reaction(reaction_id, rate, values))

    def add_rate_rule(self, rate_rule_id, val, rate_rule_min, rate_rule_max, isres):
        self.rate_rules.append(
            RateRule(rate_rule_id, val, rate_rule_min, rate_rule_max, isres)
        )

    def get_compartment_ids(self):
        return [i.id for i in self.compartments]

    def is_valid(self):
        compartments_ids = self.get_compartment_ids()

        species_compartments = [i.compartment for i in self.species]

        for count, cmp in enumerate(species_compartments):
            if cmp not in compartments_ids:
                return (
                    f"Error in species {count + 1} compartments. "
                    f"Compartment {cmp} is not present in the Compartments definition."
                )

        return 0

    def __str__(self):
        res = "Species: \n\n"

        for count, i in enumerate(self.species):
            res += f"Specie {count+1}:\n"
            res += i.__str__()

        res += "\n\n\nCompartments: \n\n"

        for count, i in enumerate(self.compartments):
            res += f"Compartment {count+1}:\n"
            res += i.__str__()

        res += "\n\n\nParameters: \n\n"

        for count, i in enumerate(self.parameters):
            res += f"Parameter {count+1}:\n"
            res += i.__str__()

        res += "\n\n\nAssignment Rules: \n\n"

        for count, i in enumerate(self.assignment_rules):
            res += f"Assignment Rule {count+1}:\n"
            res += i.__str__()

        res += "\n\n\nEvents: \n\n"

        for count, i in enumerate(self.events):
            res += f"Event {count+1}:\n"
            res += i.__str__()

        res += "\n\n\nReactions: \n\n"

        for count, i in enumerate(self.reactions):
            res += f"Reaction {count+1}:\n"
            res += i.__str__()

        res += "\n\n\nRate Rules: \n\n"

        for count, i in enumerate(self.rate_rules):
            res += f"Rate Rule {count+1}:\n"
            res += i.__str__()

        return res



from xml.etree import ElementTree as ET
import sys
from xml.etree.ElementTree import tostring
from libsbml import *

try:
    from xml_validator import validate_sbml
except:
    from sbml2hyb.xml_validator import validate_sbml


def translate_mathml(xml):
    """
    Translates the MathML formula to text
    :param xml: The MathML formula
    :return:
    """
    math = readMathMLFromString(xml)
    return formulaToString(math)


def translate_rule_math(rule):
    """
    Translates a MathML rule and returns it as text
    :param rule: The MathML formula
    :return:
    """
    if rule.isSetMath():
        formula = formulaToString(rule.getMath())

        if len(rule.getVariable()) > 0:
            return formula
        else:
            return formula + " = 0"
    else:
        return "Undefined"


def translate_reaction_math(reaction):
    """
    Translates a MathML reaction and returns it as text
    :param reaction: The MathML formula
    :return:
    """
    if reaction.isSetKineticLaw():
        kl = reaction.getKineticLaw()
        if kl.isSetMath():
            formula = formulaToString(kl.getMath())
            return formula
        else:
            return "Undefined"
    else:
        return "Undefined"


def translate_event_assignment_math(event_assignement):
    """
    Translates a MathML event assignment and returns it as text
    :param event_assignement: The MathML formula
    :return:
    """
    if event_assignement.isSetMath():
        formula = formulaToString(event_assignement.getMath())
        return formula
    else:
        return "Undefined"


def translate_event_delay_math(event):
    """
    Translates a MathML event delay and returns it as text
    :param event: The MathML formula
    :return:
    """
    if event.isSetDelay():
        formula = formulaToString(event.getDelay().getMath())
        return formula
    else:
        return "Undefined"


def translate_event_trigger_math(event):
    """
    Translates a MathML event trigger and returns it as text
    :param event: The MathML formula
    :return:
    """
    if event.isSetTrigger():
        formula = formulaToString(event.getTrigger().getMath())
        return formula
    else:
        return "Undefined"


def write_species(root, namespace, filename):
    """
    Analyzes the <species> elements in the <listOfSpecies> and builds the hmod corresponding strings
    :param root: The xml root
    :param namespace: The sbml namespace
    :param filename: The name of the file
    :return: The hmod species strings
    """
    species_for_hmod = []

    species_string = "\n"
    species_string += "% Species\n"

    species = root.findall("./*/{{{}}}listOfSpecies/*".format(namespace))
    species_string += "{}.nspecies={};\n".format(filename, len(species))
    for i, specie in enumerate(species):
        count = i + 1

        # specie id
        if "id" in specie.attrib:
            species_for_hmod.append("{}".format(specie.attrib["id"]))
            species_string += '{}.species({}).id="{}";\n'.format(
                filename, count, specie.attrib["id"]
            )

        # specie val
        if "initialConcentration" in specie.attrib:
            species_string += "{}.species({}).val={};\n".format(
                filename, count, specie.attrib["initialConcentration"]
            )

        # specie compartment
        if "compartment" in specie.attrib:
            species_string += '{}.species({}).compartment="{}";\n'.format(
                filename, count, specie.attrib["compartment"]
            )

        # specie fixed
        if "constant" in specie.attrib:
            species_string += "{}.species({}).fixed={};\n".format(
                filename, count, 0 if specie.attrib["constant"] == "false" else 1
            )
        else:
            species_string += "{}.species({}).fixed={};\n".format(filename, count, 0)

        # specie min
        species_string += "{}.species({}).min={};\n".format(filename, count, 0)

        # specie max
        species_string += "{}.species({}).max={};\n".format(filename, count, 15)

        # specie isres
        species_string += "{}.species({}).isres={};\n".format(filename, count, 1)

    return [species_string, species_for_hmod]


def write_compartments(root, namespace, filename):
    """
    Analyzes the <compartment> elements in the <listOfCompartments> and builds the hmod corresponding strings
    :param root: The xml root
    :param namespace: The sbml namespace
    :param filename: The name of the file
    :return: The hmod compartments strings
    """
    compartments_string = "\n"
    compartments_string += "% Compartments\n"
    compartments = root.findall("./*/{{{}}}listOfCompartments/*".format(namespace))
    compartments_string += "{}.ncompartments={};\n".format(filename, len(compartments))

    for i, compartment in enumerate(compartments):
        count = i + 1

        # compartment id
        if "id" in compartment.attrib:
            compartments_string += '{}.compartments({}).id="{}";\n'.format(
                filename, count, compartment.attrib["id"]
            )

        # compartment val
        if "size" in compartment.attrib:
            compartments_string += "{}.compartments({}).val={};\n".format(
                filename, count, compartment.attrib["size"]
            )

        # compartment fixed
        if "constant" in compartment.attrib:
            compartments_string += "{}.compartments({}).fixed={};\n".format(
                filename, count, 0 if compartment.attrib["constant"] == "false" else 1
            )
        else:
            compartments_string += "{}.compartments({}).fixed={};\n".format(
                filename, count, 0
            )

        # compartment min
        compartments_string += "{}.compartments({}).min={};\n".format(
            filename, count, 0
        )

        # compartment max
        compartments_string += "{}.compartments({}).max={};\n".format(
            filename, count, 15
        )

        # compartment isres
        compartments_string += "{}.compartments({}).isres={};\n".format(
            filename, count, 0
        )

    return compartments_string


def write_parameters(root, namespace, filename):
    """
    Analyzes the <parameter> elements in the <listOfParameters> and builds the hmod corresponding strings
    :param root: The xml root
    :param namespace: The sbml namespace
    :param filename: The name of the file
    :return: The hmod parameters strings
    """
    parameters_for_hmod = []
    parameters_string = "\n"
    parameters_string += "% Parameters\n"
    parameters = root.findall("./*/{{{}}}listOfParameters/*/[@id]".format(namespace))
    local_parameters = root.findall(
        "./*/{{{}}}listOfReactions/*/{{{}}}kineticLaw/{{{}}}listOfParameters/*".format(
            namespace, namespace, namespace
        )
    )
    local_parameters += root.findall(
        "./*/{{{}}}listOfReactions/*/{{{}}}kineticLaw/{{{}}}listOfLocalParameters/*".format(
            namespace, namespace, namespace
        )
    )
    aux = []
    for i, local_parameter in enumerate(local_parameters):
        can_be_inputted = True
        for j in range(len(aux)):
            if "id" in aux[j].attrib and "id" in local_parameter.attrib:
                if aux[j].attrib["id"] == local_parameter.attrib["id"]:
                    can_be_inputted = False
                    break
        if can_be_inputted:
            aux.append(local_parameter)

    local_parameters = aux
    number_of_local_parameters = 0
    for i, local_parameter in enumerate(local_parameters):
        # if parameter.attrib["value"] == "0":
        #     continue

        # Check if parameter is Global
        is_local = True
        for j, parameter in enumerate(parameters):
            if "id" in local_parameter.attrib and "id" in parameter.attrib:
                if parameter.attrib["id"] == local_parameter.attrib["id"]:
                    is_local = False
                    break
        if not is_local:
            continue
        number_of_local_parameters += 1

    parameters_string += "{}.nparameters={};\n".format(
        filename, number_of_local_parameters + len(parameters)
    )
    param_counter = 0

    for i, parameter in enumerate(parameters):
        # if parameter.attrib["value"] == "0":
        #     continue

        param_counter += 1
        # parameter id
        if "id" in parameter.attrib:
            parameters_for_hmod.append("{}".format(parameter.attrib["id"]))
            parameters_string += '{}.parameters({}).id="{}";\n'.format(
                filename, param_counter, parameter.attrib["id"]
            )

        # parameter val
        if "value" in parameter.attrib:
            parameters_string += "{}.parameters({}).val={};\n".format(
                filename, param_counter, parameter.attrib["value"]
            )

        # parameter fixed
        if "constant" in parameter.attrib:
            parameters_string += "{}.parameters({}).fixed={};\n".format(
                filename,
                param_counter,
                0 if parameter.attrib["constant"] == "false" else 1,
            )
        else:
            parameters_string += "{}.parameters({}).fixed={};\n".format(
                filename, param_counter, 0
            )

        # parameter reaction
        parameters_string += '{}.parameters({}).reaction="{}";\n'.format(
            filename, param_counter, "global"
        )

    for local_parameter in local_parameters:

        # Check if parameter is in Local
        is_global = False
        for j, global_parameter in enumerate(parameters):
            if "id" in local_parameter.attrib and "id" in global_parameter.attrib:
                if local_parameter.attrib["id"] == global_parameter.attrib["id"]:
                    is_global = True
                    break
        if is_global:
            continue

        param_counter += 1
        # local parameter id
        if "id" in local_parameter.attrib:
            parameters_for_hmod.append("{}".format(local_parameter.attrib["id"]))
            parameters_string += '{}.parameters({}).id="{}";\n'.format(
                filename, param_counter, local_parameter.attrib["id"]
            )

        # local parameter val
        if "value" in local_parameter.attrib:
            parameters_string += "{}.parameters({}).val={};\n".format(
                filename, param_counter, local_parameter.attrib["value"]
            )

        # parameter fixed
        if "constant" in local_parameter.attrib:
            parameters_string += "{}.parameters({}).fixed={};\n".format(
                filename,
                param_counter,
                0 if local_parameter.attrib["constant"] == "false" else 1,
            )
        else:
            parameters_string += "{}.parameters({}).fixed={};\n".format(
                filename, param_counter, 0
            )

        # local parameter reaction
        parameters_string += '{}.parameters({}).reaction="{}";\n'.format(
            filename, param_counter, "local"
        )

    return [parameters_string, parameters_for_hmod]


def write_assignment_rules(root, namespace, filename):
    """
    Analyzes the <assignmentRule> elements in the <listOfRules> and builds the hmod corresponding strings
    :param root: The xml root
    :param namespace: The sbml namespace
    :param filename: The name of the file
    :return: The hmod assignmentRule strings
    """

    assignment_rules_string = "\n"
    assignment_rules_string += "% Assignment rules\n"
    ruleAss = root.findall(
        "./*/{{{}}}listOfRules/{{{}}}assignmentRule".format(namespace, namespace)
    )
    assignment_rules_string += "{}.nruleAss={};\n".format(filename, len(ruleAss))

    for i, assignment_rule in enumerate(ruleAss):
        count = i + 1

        # assignment_rule id
        if "variable" in assignment_rule.attrib:
            assignment_rules_string += '{}.ruleAss({}).id="{}";\n'.format(
                filename, count, assignment_rule.attrib["variable"]
            )

        # assignment_rule val
        # val = translate_mathml(assignment_rule.find(
        #     './{http://www.w3.org/1998/Math/MathML}math/*'))
        rule = tostring(
            assignment_rule.find("./{http://www.w3.org/1998/Math/MathML}math"),
            encoding="unicode",
            method="xml",
        )
        val = translate_mathml(rule)
        assignment_rules_string += '{}.ruleAss({}).val="{}";\n'.format(
            filename, count, val
        )

    return assignment_rules_string


def write_events(root, namespace, filename):
    """
    Analyzes the <event> elements in the <listOfEvents> and builds the hmod corresponding strings
    :param root: The xml root
    :param namespace: The sbml namespace
    :param filename: The name of the file
    :return: The hmod events strings
    """

    events_string = "\n"
    events_string += "% Events\n"
    events = root.findall("./*/{{{}}}listOfEvents/*".format(namespace))
    events_string += "{}.nevents={};\n".format(filename, len(events))

    for i, event in enumerate(events):
        count = i + 1

        # event id
        events_string += '{}.events({}).id="{}";\n'.format(
            filename, count, "Controller"
        )

        # event trigger
        # trigger = translate_mathml(event.find(
        #     './{{{}}}trigger/{{http://www.w3.org/1998/Math/MathML}}math/*'.format(namespace)))
        math_trigger = tostring(
            event.find(
                "./{{{}}}trigger/{{http://www.w3.org/1998/Math/MathML}}math".format(
                    namespace
                )
            ),
            encoding="unicode",
            method="xml",
        )
        trigger = translate_mathml(math_trigger)

        events_string += '{}.events({}).trigger="{}";\n'.format(
            filename, count, trigger
        )

        # event delay
        math_delay = tostring(
            event.find(
                "./{{{}}}delay/{{http://www.w3.org/1998/Math/MathML}}math".format(
                    namespace
                )
            ),
            encoding="unicode",
            method="xml",
        )
        delay = translate_mathml(math_delay)
        events_string += '{}.events({}).delay="{}";\n'.format(filename, count, delay)

        event_assignements = event.findall(
            "./{{{}}}listOfEventAssignments/*".format(namespace)
        )

        for j, event_assignement in enumerate(event_assignements):
            ass_count = j + 1

            # event assignement var
            if "variable" in event_assignement.attrib:
                events_string += '{}.events({}).var({})="{}";\n'.format(
                    filename, count, ass_count, event_assignement.attrib["variable"]
                )

            # event assignement var
            # var = translate_mathml(event_assignement.find(
            #     './{http://www.w3.org/1998/Math/MathML}math/*'))
            math_var = tostring(
                event_assignement.find("./{http://www.w3.org/1998/Math/MathML}math"),
                encoding="unicode",
                method="xml",
            )
            var = translate_mathml(math_var)
            events_string += '{}.events({}).var({})="{}";\n'.format(
                filename, count, ass_count, var
            )

    return events_string


def write_reactions(root, namespace, filename):
    """
    Analyzes the <reaction> elements in the <listOfReactions> and builds the hmod corresponding strings
    :param root: The xml root
    :param namespace: The sbml namespace
    :param filename: The name of the file
    :return: The hmod reaction strings
    """

    reactions_string = "\n"
    reactions_string += "% Reactions\n"
    reactions = root.findall("./*/{{{}}}listOfReactions/*".format(namespace))
    reactions_string += "{}.nreaction={};\n".format(filename, len(reactions))

    species = root.findall("./*/{{{}}}listOfSpecies/*".format(namespace))
    species_list = []
    for specie in species:
        if "id" in specie.attrib:
            species_list += [specie.attrib["id"]]

    for i, reaction in enumerate(reactions):
        count = i + 1
        # reaction id
        if "id" in reaction.attrib:
            reactions_string += '{}.reaction({}).id="{}";\n'.format(
                filename, count, reaction.attrib["id"]
            )

        # reaction rate
        # rate = translate_mathml(reaction.find(
        #     './{{{}}}kineticLaw/{{http://www.w3.org/1998/Math/MathML}}math/*'.format(namespace)))
        # reactions_string += ('{}.reaction({}).rate="{}";\n'.format(filename, count, rate))
        # rate = translate_reaction_math(model.getReaction(i))
        math_rate = tostring(
            reaction.find(
                "./{{{}}}kineticLaw/{{http://www.w3.org/1998/Math/MathML}}math".format(
                    namespace
                )
            ),
            encoding="unicode",
            method="xml",
        )
        rate = translate_mathml(math_rate)
        reactions_string += '{}.reaction({}).rate="{}";\n'.format(filename, count, rate)

        values = {k: "0" for k in species_list}
        reactants_list = reaction.findall("./{{{}}}listOfReactants/*".format(namespace))
        reactants_dict = {
            k.attrib["species"]: k.attrib["stoichiometry"]
            if "stoichiometry" in k.attrib
            else "1"
            for k in reactants_list
        }

        products_list = reaction.findall("./{{{}}}listOfProducts/*".format(namespace))
        products_dict = {
            k.attrib["species"]: k.attrib["stoichiometry"]
            if "stoichiometry" in k.attrib
            else "1"
            for k in products_list
        }

        for j in reactants_dict:
            values[j] = "-" + reactants_dict[j] if reactants_dict[j] != "0" else "0"
        for j in products_dict:
            values[j] = products_dict[j]

        # reaction values
        reactions_string += '{}.reaction({}).{}="{}";\n'.format(
            filename, count, reaction.attrib["id"], list(values.values())
        )

    return reactions_string


def write_rate_rules(root, namespace, filename):
    """
    Analyzes the <rateRule> elements in the <listOfRules> and builds the hmod corresponding strings
    :param root: The xml root
    :param namespace: The sbml namespace
    :param filename: The name of the file
    :return: The hmod rateRules strings
    """

    rate_rules_string = "\n"
    rate_rules_string += "% Rate rules\n"
    rateRules = root.findall(
        "./*/{{{}}}listOfRules/{{{}}}rateRule".format(namespace, namespace)
    )
    rate_rules_string += "{}.nraterules={};\n".format(filename, len(rateRules))

    for i, rate_rule in enumerate(rateRules):
        count = i + 1
        # assignment_rule id
        if "variable" in rate_rule.attrib:
            rate_rules_string += '{}.raterules({}).id="{}";\n'.format(
                filename, count, rate_rule.attrib["variable"]
            )

        # assignment_rule val
        # val = translate_mathml(rate_rule.find(
        #     './{http://www.w3.org/1998/Math/MathML}math/*'))

        rule = tostring(
            rate_rule.find("./{http://www.w3.org/1998/Math/MathML}math"),
            encoding="unicode",
            method="xml",
        )
        val = translate_mathml(rule)

        rate_rules_string += '{}.raterules({}).val="{}";\n'.format(filename, count, val)

        # compartment min
        rate_rules_string += "{}.raterules({}).min={};\n".format(filename, count, 0)

        # compartment max
        rate_rules_string += "{}.raterules({}).max={};\n".format(filename, count, 1.5)

        # compartment isres
        rate_rules_string += "{}.raterules({}).isres={};\n".format(filename, count, 0)

    return rate_rules_string


def main(filename, file_string):
    a = validate_sbml(file_string)
    if a != 0:
        return [
            "The SBML file is not valid. See the reason(s) below: \n\n\n" + a,
            None,
            None,
        ]

    # parsing the xml
    tree = ET.fromstring(file_string)  # open(filename))
    # getting the xml root
    root = tree

    # getting the sbml namespace
    namespace = root.tag.replace("{", "").replace("}sbml", "")

    # building the hmod filename, wich is the same as the xml one
    if ".xml" in filename:
        hmod_filename = filename.split("/")[-1].replace(".xml", "")
    else:
        hmod_filename = filename.split("/")[-1].replace(".hmod", "")

    # building the hmod file string
    f = "% --------------------------------------------------------------------------\n"
    # f += ("% hybrid model definition data structure\n")
    f += "% hmod model structure\n"
    res = write_species(root, namespace, hmod_filename)
    species = res[1]
    f += res[0]
    f += write_compartments(root, namespace, hmod_filename)
    res = write_parameters(root, namespace, hmod_filename)
    parameters = res[1]
    f += res[0]
    f += write_assignment_rules(root, namespace, hmod_filename)
    f += write_events(root, namespace, hmod_filename)
    f += write_reactions(root, namespace, hmod_filename)
    f += write_rate_rules(root, namespace, hmod_filename)
    f += "\n"
    f += "\n"
    f += "end"

    return [f, species, parameters]


if __name__ == "__main__":
    # calling main function
    if len(sys.argv) != 2:
        print("Usage: python xml_parser.py 'filename'")
    else:
        file = sys.argv[1]
        main(file)



import os
import sys
from xml.etree import ElementTree as ET
from xml.dom import minidom
from xml.etree.ElementTree import fromstring, register_namespace
import ast
from libsbml import *

try:
    from Hmod import Hmod
    from hmod_validator import validate_hmod
except:
    from sbml2hyb.Hmod import Hmod
    from sbml2hyb.hmod_validator import validate_hmod

meta_id = 0


def build_xml_string(sbml):
    """
    Builds the sbml string of the file
    :param sbml: The xml root element
    :return:
    """

    # builds the string file with proper identation
    xmlstr = minidom.parseString(
        ET.tostring(sbml, xml_declaration=True, encoding="UTF-8")
    ).toprettyxml(indent="\t", encoding="UTF-8")

    # necessary adjustments to make the produced file cleaner
    string_to_write = (
        xmlstr.decode("utf-8")
        .replace(' xmlns:math="http://www.w3.org/1998/Math/MathML"', "")
        .replace("<math:math", '<math xmlns="http://www.w3.org/1998/Math/MathML"')
        .replace("math:", "")
        .replace(
            "<ci> time </ci>",
            '<csymbol encoding="text" definitionURL="http://www.sbml.org/sbml/symbols/time"> time </csymbol>',
        )
        .replace(
            "<ci> t </ci>",
            '<csymbol encoding="text" definitionURL="http://www.sbml.org/sbml/symbols/time"> time </csymbol>',
        )
    )

    # removing empty lines
    string_to_write = "".join(
        [s for s in string_to_write.strip().splitlines(True) if s.strip()]
    )
    return string_to_write


def get_parameters(lines):
    """
    Returns the global parameters in hmod strings
    :param lines: The file string lines
    :return:
    """
    parameters = []
    for i in lines:
        if "parameters(" in i:
            parameters += [i]
    return parameters


def get_local_parameters(lines):
    """
    Returns the global parameters in hmod strings
    :param lines: The file string lines
    :return:
    """
    parameters = []
    for i in lines:
        if "parameters(" in i:
            parameters += [i]
        if "global" in i:
            parameters = parameters[: len(parameters) - 4]
    return parameters


def get_global_parameters(lines):
    """
    Returns the global parameters in hmod strings
    :param lines: The file string lines
    :return:
    """
    parameters = []
    for i in lines:
        if "parameters(" in i:
            parameters += [i]
        if "local" in i:
            parameters = parameters[: len(parameters) - 4]
    return parameters


def get_local_parameters_dict(lines):
    """
    Returns a dict with {"parameter": parameter value}
    :param lines: The file string lines
    :return:
    """
    global_parameters = get_global_parameters(lines)

    par = [
        get_value_from_string(i)
        for i in lines
        if "parameters(" in i and "id" in i and i not in global_parameters
    ]
    par_values = [
        get_value_from_string(i)
        for i in lines
        if "parameters(" in i and "val" in i and i not in global_parameters
    ]
    param_dict = {param: par_values[i] for i, param in enumerate(par)}
    return param_dict


def get_value_from_string(string):
    """
    Returns the desired attribute value from an hmod string
    :param string: The whole string
    :return:
    """
    return str(string.split("=")[-1].replace('"', "").replace(";", "").replace("'", ""))


def translateInfix(formula):
    """
    Translates the text formula to MathML
    :param formula:
    :return: The MathML
    """
    math = parseFormula(formula)
    return writeMathMLToString(math)


def get_attribute_list(lines, category, category_number):
    """
    Returns the attribute list for a certain category
    Example for an assignment rule:

    Park_SBML.nruleAss=1;
    Park_SBML.ruleAss(1).id="D";
    Park_SBML.ruleAss(1).val="Feed / V";

    Is converted to the list ["D", "Feed / V"]

    :param lines: File string lines
    :param category: The string representing the hmod category
    :param category_number: The number of attributes the category has
    :return:
    """
    # collect every string related to the category
    attributes_list = [get_value_from_string(i) for i in lines if category in i]

    # divides all category attributes into multiple lists to process each one separately
    #
    # For compartments list:
    # for example, for a file with 2 compartments, x would be:
    # [
    # ['V', '1', '0', '15', '0'],
    # ['H', '1', '0', '15', '0'],
    # ]
    # Being the order: [id, val, fixed, min, max]
    x = [
        attributes_list[i : i + category_number]
        for i in range(0, len(attributes_list), category_number)
    ]
    return x


def store_compartments(compartments, list_of_compartments):
    """
    Produces the <listOfCompartments>
    :param compartments: A list with all compartments attributes
    :param list_of_compartments: The xml element
    """
    global meta_id

    # processing each compartment
    for i in compartments:
        compartment = ET.SubElement(
            list_of_compartments,
            "compartment",
            metaid="metaid_" + str(meta_id),
            id=i[0],
            size=i[1],
            constant="false" if i[2] == "0" else "true",
        )
        meta_id += 1


def store_species(species, list_of_species):
    """
    Produces the <listOfSpecies>
    :param species: A list with all species attributes
    :param list_of_species: The xml element
    """
    global meta_id

    # processing each species
    for i in species:
        specie = ET.SubElement(
            list_of_species,
            "species",
            metaid="metaid_" + str(meta_id),
            id=i[0],
            compartment=i[2],
            initialConcentration=i[1],
            constant="false" if i[3] == "0" else "true",
        )
        meta_id += 1


def store_parameters(lines, list_of_parameters):
    """
    Produces the <listOfParameters>
    :param lines: The file string lines
    :param list_of_parameters: The xml element
    :return id list of parameters
    """
    global meta_id

    # collect every string related to parameters
    parameters = get_global_parameters(lines)  # + get_local_parameters(lines)

    # the number of attributes that a parameter has in hmod
    parameters_number = 4

    # divides all parameters into multiple lists to process each one separately
    x = [
        parameters[i : i + parameters_number]
        for i in range(0, len(parameters), parameters_number)
    ]

    list_of_ids = []
    # processing each parameter
    for i in x:
        cur_id = get_value_from_string(i[0])
        list_of_ids.append(cur_id)
        parameter = ET.SubElement(
            list_of_parameters,
            "parameter",
            metaid="metaid_" + str(meta_id),
            id=cur_id,
            value=get_value_from_string(i[1]),
            constant="false" if get_value_from_string(i[2]) == "0" else "true",
        )
        meta_id += 1

    local_params = get_local_parameters_dict(lines)
    for key in local_params.keys():
        list_of_ids.append(key)

    return list_of_ids


def store_rate_rules(rate_rules, list_of_rate_rules):
    """
    Produces all the <rateRule> elements in a <listOfRules>
    :param rate_rules: A list with all rate rules attributes
    :param list_of_rate_rules: The xml element
    """
    global meta_id

    # processing each rate rule
    for i in rate_rules:
        compartment = ET.SubElement(
            list_of_rate_rules,
            "rateRule",
            metaid="metaid_" + str(meta_id),
            variable=i[0],
        )

        # builds the MathML
        equation = i[1]
        eq = fromstring(
            translateInfix(equation).replace(
                '<?xml version="1.0" encoding="UTF-8"?>', ""
            )
        )
        compartment.append(eq)
        meta_id += 1


def store_assignment_rules(assignment_rules, list_of_assignment_rules):
    """
    Produces all the <assignmentRule> elements in a <listOfRules>
    :param assignment_rules: A list with all assignment rules attributes
    :param list_of_assignment_rules: The xml element
    """

    global meta_id

    # processing each assignment rule
    for i in assignment_rules:
        assignment_rule = ET.SubElement(
            list_of_assignment_rules,
            "assignmentRule",
            metaid="metaid_" + str(meta_id),
            variable=i[0],
        )

        # builds the MathML
        equation = i[1]
        eq = fromstring(
            translateInfix(equation).replace(
                '<?xml version="1.0" encoding="UTF-8"?>', ""
            )
        )
        assignment_rule.append(eq)
        meta_id += 1


def store_reactions(lines, list_of_reactions, species_list):
    """
    Produces the <listOfReactions>
    :param lines: The file string lines
    :param list_of_reactions: The xml element
    :param species_list: The string list of all species
    """
    global meta_id

    # collect every string related to reactions
    reactions = [i for i in lines if "reaction(" in i]

    # the number of attributes that a reaction has in hmod
    reactions_number = 3

    # divides all reactions into multiple lists to process each one separately
    x = [
        reactions[i : i + reactions_number]
        for i in range(0, len(reactions), reactions_number)
    ]

    # processing each reaction
    for i in x:

        reaction = ET.SubElement(
            list_of_reactions,
            "reaction",
            metaid="metaid_" + str(meta_id),
            id=get_value_from_string(i[0]),
        )
        meta_id += 1

        # Transforming a string containing a list with the stoichiometric values into a proper list of strings
        # e.g. The string "['1', '2', '3', '4']" into the list ['1', '2', '3', '4']
        x = ast.literal_eval(get_value_from_string(i[2]))

        # converting the stoichiometric string values into floats
        species_values = [float(k) for k in x]

        # list of products dict, with the key being the specie index in the spceies_list, and the corresponding
        # stoichiometric value
        products = {}

        # list of reactants dict, with the key being the specie index in the spceies_list, and the corresponding
        # stoichiometric value
        reactants = {}

        # filling the products and reactants dicts
        # if the specie as a stoichiometric value equal to 0 it does not participate in the reaction
        # if the specie as a stoichiometric value lesser than 0 it is a reactant
        # if the specie as a stoichiometric value greater than 0 it is a product
        for j, value in enumerate(species_values):
            if value < 0:
                reactants[j] = abs(value)
            elif value > 0:
                products[j] = value

        # if there are any reactants, builds the <listOfReactants>
        if len(reactants) > 0:
            listOfReactants = ET.SubElement(reaction, "listOfReactants")
            # building each <speciesReference>
            for key, value in reactants.items():
                speciesReference = ET.SubElement(
                    listOfReactants,
                    "speciesReference",
                    metaid="metaid_" + str(meta_id),
                    species=species_list[key],
                    stoichiometry=str(value),
                )
                meta_id += 1

        # if there are any products, builds the <listOfProducts>
        if len(products) > 0:
            listOfProducts = ET.SubElement(reaction, "listOfProducts")
            # building each <speciesReference>
            for key, value in products.items():
                speciesReference = ET.SubElement(
                    listOfProducts,
                    "speciesReference",
                    metaid="metaid_" + str(meta_id),
                    species=species_list[key],
                    stoichiometry=str(value),
                )
                meta_id += 1

        # Builds the MathML
        equation = get_value_from_string(i[1])
        eq = fromstring(
            translateInfix(equation).replace(
                '<?xml version="1.0" encoding="UTF-8"?>', ""
            )
        )

        # searching for all <modifierSpeciesReference> elements of <listOfModifiers>
        # if it is a product or reactant, then is added
        # if it is referenced in the MathML it is added, as long as it is a species
        modifiers = [
            i.text.strip()
            for i in eq.findall(".//{http://www.w3.org/1998/Math/MathML}ci")
            if i.text.strip() in species_list
        ]
        modifiers = [m for m in modifiers if m != ""]
        modifiers += [species_list[i] for i in products]
        modifiers += [species_list[i] for i in reactants]
        modifiers = list(dict.fromkeys(modifiers))

        # building the <listOfModifiers> xml element
        listOfModifiers = ET.SubElement(reaction, "listOfModifiers")

        # processing each modifier
        for modifier in modifiers:
            modifierSpeciesReference = ET.SubElement(
                listOfModifiers,
                "modifierSpeciesReference",
                metaid="metaid_" + str(meta_id),
                species=modifier,
            )
            meta_id += 1

        # finally appends the MathMl to the <kineticLaw>
        kineticLaw = ET.SubElement(reaction, "kineticLaw")
        kineticLaw.append(eq)

        # finding the local parameters of a kinetic law
        # store all the elements that are inside a <ci> and are local parameters
        local_parameters = get_local_parameters_dict(lines)
        parameters = [
            i.text.strip()
            for i in eq.findall(".//{http://www.w3.org/1998/Math/MathML}ci")
            if i.text.strip() in local_parameters
        ]

        # removing duplicates
        parameters = list(set(parameters))

        # remove global
        global_parameters = get_global_parameters(lines)
        aux = [
            global_parameters[i : i + 4] for i in range(0, len(global_parameters), 4)
        ]
        final_parameters = []
        for parameter in parameters:
            is_global = False
            for i in aux:
                cur_id = get_value_from_string(i[0])
                if parameter == cur_id:
                    is_global = True
                    break
            if not is_global:
                final_parameters.append(parameter)

        parameters = final_parameters
        # adding the parameters
        if len(parameters) > 0:
            listOfParameters = ET.SubElement(kineticLaw, "listOfParameters")
            for param in parameters:
                parameter = ET.SubElement(
                    listOfParameters,
                    "parameter",
                    metaid="metaid_" + str(meta_id),
                    id=param,
                    value=local_parameters[param],
                )
                meta_id += 1


def store_events(events, list_of_events):
    """
    Produces the <listOfEvents>
    :param events: A list with all events attributes
    :param list_of_events: The xml element
    """
    global meta_id

    # # collect every string related to reactions
    # events = [i for i in lines if "events(" in i]
    #
    # # the number of attributes that an event has in hmod
    # events_number = 5
    #
    # # divides all events into multiple lists to process each one separately
    # x = [events[i:i + events_number] for i in range(0, len(events), events_number)]

    # processing each event
    for i in events:
        event = ET.SubElement(list_of_events, "event", metaid="metaid_" + str(meta_id))

        # builds the <trigger> element
        trigger = ET.SubElement(event, "trigger")
        trigger_equation = get_value_from_string(i[1])
        eq = fromstring(
            translateInfix(trigger_equation).replace(
                '<?xml version="1.0" encoding="UTF-8"?>', ""
            )
        )
        trigger.append(eq)

        # builds the <delay> element
        delay = ET.SubElement(event, "delay")
        delay_equation = get_value_from_string(i[2])
        eq = fromstring(
            translateInfix(delay_equation).replace(
                '<?xml version="1.0" encoding="UTF-8"?>', ""
            )
        )
        delay.append(eq)

        meta_id += 1

        # builds the <listOfEventAssignments> for this event
        store_event_assignments(event, i[3:])


def store_event_assignments(event, event_assignments):
    """
    Produces the <listOfEventAssignments> for an <event>
    :param event: The event
    :param event_assignments: The string lines containing the event assignments attributes
    :return:
    """
    # building the <listOfEventAssignments> xml element
    listOfEventAssignments = ET.SubElement(event, "listOfEventAssignments")

    # the number of attributes that an event assignment has in hmod
    event_assignments_number = 2

    # divides all event assignments into multiple lists to process each one separately
    x = [
        event_assignments[i : i + event_assignments_number]
        for i in range(0, len(event_assignments), event_assignments_number)
    ]

    # processing each event assignment
    for i in x:
        event_assignment = ET.SubElement(
            listOfEventAssignments,
            "eventAssignment",
            metaid="metaid_" + str(meta_id),
            variable=get_value_from_string(i[0]),
        )

        # builds the MathML
        event_assignment_equation = get_value_from_string(i[1])
        eq = fromstring(
            translateInfix(event_assignment_equation).replace(
                '<?xml version="1.0" encoding="UTF-8"?>', ""
            )
        )
        event_assignment.append(eq)


def main(filename, doc):

    lines = doc.split("\n")

    # Strips the newline character
    for count, line in enumerate(lines):
        lines[count] = line.strip()
        count += 1
    hmod_valid = validate_hmod(lines)
    if hmod_valid != 0:
        return [hmod_valid, None, None]

    # resgister the mathml namespace
    register_namespace("math", "http://www.w3.org/1998/Math/MathML")

    # build the <sbml>
    sbml = ET.Element(
        "sbml", xmlns="http://www.sbml.org/sbml/level2/version4", level="2", version="4"
    )

    # build the <model>
    model = ET.SubElement(
        sbml,
        "model",
        id=(
            filename.split("/")[-1].replace(".hmod", "")
            if ".hmod" in filename
            else filename.split("/")[-1].replace(".xml", "")
        ),
    )

    # builds the <listOfCompartments>
    compartments_number = 6
    compartments = get_attribute_list(lines, "compartments(", compartments_number)
    if compartments:
        listOfCompartments = ET.SubElement(model, "listOfCompartments")
        store_compartments(compartments, listOfCompartments)

    # builds the <listOfSpecies>
    species_number = 7
    species = get_attribute_list(lines, "species(", species_number)
    if species:
        listOfSpecies = ET.SubElement(model, "listOfSpecies")
        store_species(species, listOfSpecies)

    # builds the <listOfParameters>
    listOfParameters = ET.SubElement(model, "listOfParameters")
    p = store_parameters(lines, listOfParameters)

    # builds the <listOfRules>

    rate_rules_number = 5
    rate_rules = get_attribute_list(lines, "raterules(", rate_rules_number)
    assignment_rules_number = 2
    assignment_rules = get_attribute_list(lines, "ruleAss(", assignment_rules_number)

    if rate_rules or assignment_rules:
        listOfRules = ET.SubElement(model, "listOfRules")

    if rate_rules:
        store_rate_rules(rate_rules, listOfRules)

    if assignment_rules:
        store_assignment_rules(assignment_rules, listOfRules)

    # builds the <listOfReactions>
    listOfReactions = ET.SubElement(model, "listOfReactions")
    species = [i.attrib["id"] for i in sbml.findall("./*/listOfSpecies/*")]
    store_reactions(lines, listOfReactions, species)

    # builds the <listOfEvents>
    events_number = 5
    events = get_attribute_list(lines, "events(", events_number)
    if events:
        listOfEvents = ET.SubElement(model, "listOfEvents")
        store_events(events, listOfEvents)

    global meta_id
    meta_id = 0

    return [build_xml_string(sbml), species, p]


if __name__ == "__main__":
    # calling main function
    if len(sys.argv) != 2:
        print("Usage: python xml_parser.py 'filename'")
    else:
        file = sys.argv[1]
        main(file)



from distutils.core import setup
from setuptools import find_packages
import os

# Optional project description in README.md:
current_directory = os.path.dirname(os.path.abspath(__file__))

try:
    with open(os.path.join(current_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
except Exception:
    long_description = ''
setup(

# Project name: 
name='sbml2hyb',

# Packages to include in the distribution: 
package_dir={"":"src"},

# Packages to include in the distribution: 
packages=find_packages(where='src'),

# Project version number:
version='1.0.3',

# Overall description
description="sbml2hyb is a Pyton tool for SBML compatible hybrid modelling.",

# Long description of your library: 
long_description=long_description,
long_description_content_type='text/markdown',

# Your name: 
author='Rafael Costa',

# Your email address:
author_email='rs.costa@fct.unl.pt',

# Link to your github repository or website: 
url='https://github.com/r-costa/sbml2hyb',

# Download Link from where the project can be downloaded from:
download_url='https://github.com/r-costa/sbml2hyb',

# List of keywords
keywords=[
    "python","converter", "gui", "neural-network", "interface", "systems-biology", "hybrid-model", "keras-tensorflow", "sbml-model"
],

# List project dependencies: 
install_requires=[
'Pillow==9.0',
'tensorflow==2.10.0',
'python-libsbml==5.19.6'
],

# https://pypi.org/classifiers/ 
classifiers=[
    "Development Status :: 5 - Production/Stable",
    "Programming Language :: Python :: 3.7",
    "License :: OSI Approved :: GNU Affero General Public License v3",
    "Operating System :: OS Independent",
]
)


import re

try:
    from Hmod import Hmod
except:
    from sbml2hyb.Hmod import Hmod


def get_number(line):
    number = line.split("=")[-1].split(";")[0]
    try:
        return int(number)
    except ValueError:
        return float(number)


def get_parameter_value(parameter):
    return parameter.split("=")[-1].replace('"', "").replace("'", "").replace(";", "")


def match_patterns(category, parameter, hmod):
    """
    Check if the file lines match the correct patterns
    :param hmod:
    :param category: The category that will be analyzed
    :param parameter: The file line
    :return:
    """

    category_dict = {
        "species": [["id", "compartment"], ["val", "fixed", "min", "max", "isres"]],
        "compartments": [["id"], ["val", "fixed", "min", "max", "isres"]],
        "parameters": [["id", "reaction"], ["val", "fixed"]],
        "ruleAss": [["id", "val"], []],
        "events": [["id", "trigger", "delay", "var[(][0-9]*[)]"], []],
        "reaction": [["id", "rate"], []],
        "raterules": [["id", "val"], ["min", "max", "isres"]],
    }
    string_parameters = category_dict[category][0]
    numbers_parameters = category_dict[category][1]

    reaction_string = r"^[a-zA-Z0-9\_]*\.reaction\([0-9]*\)\.{reaction_id}=(\"|')\[[a-zA-Z0-9 \/\*\+\-\'\,\"]*\](\"|');$"

    for s in string_parameters:

        pattern_string = rf"^[a-zA-Z0-9\_]*\.{category}\([0-9]*\)\.{s}=(\"|').*(\"|');$"

        match = bool(re.match(pattern_string, parameter))
        if match:
            return True

    for s in numbers_parameters:
        pattern_number = rf"^[a-zA-Z0-9\_]*[.]{category}[(][0-9]*[)][.]{s}=-?([0-9]*[.])?[0-9]+((E|e)\-[0-9]*)?;$"

        match = bool(re.match(pattern_number, parameter))
        if match:
            return True

    return False


def match_reaction_pattern(line, expected_value):
    reaction_string = rf"^[a-zA-Z0-9\_]*\.reaction\([0-9]*\)\.{expected_value}=(\"|')\[[a-zA-Z0-9 \/\*\+\-\'\,\"\.]*\](\"|');$"

    match = bool(re.match(reaction_string, line))
    if match:
        return True
    return False


def get_parameters_from_list(attributes):
    return [i.split(".")[2].split("=")[0].split("(")[0] for i in attributes]


def get_parameter(line):
    return line.split(".")[2].split("=")[0].split("(")[0]


def get_category_lines(category, lines):
    return [i for i in lines if f".n{category}=" in i or f".{category}(" in i]


def get_category_examples(category):
    examples_dict = {
        "species": "FILENAME.nspecies=4;\n"
        'FILENAME.species(1).id="PM";\n'
        "FILENAME.species(1).val=0;\n"
        'FILENAME.species(1).compartment="V";\n'
        "FILENAME.species(1).fixed=0;\n"
        "FILENAME.species(1).min=0;\n"
        "FILENAME.species(1).max=15;\n"
        "FILENAME.species(1).isres=1;\n"
        ".....\n\n",
        "compartments": "Park_SBML.ncompartments=1;\n"
        'Park_SBML.compartments(1).id="V";'
        "\nPark_SBML.compartments(1).val=1;\n"
        "Park_SBML.compartments(1).fixed=0;\n"
        "Park_SBML.compartments(1).min=0;\n"
        "Park_SBML.compartments(1).max=15;\n"
        "Park_SBML.compartments(1).isres=0;\n"
        ".....\n\n",
        "parameters": "Park_SBML.nparameters=9;\n"
        'Park_SBML.parameters(1).id="D";\n'
        "Park_SBML.parameters(1).val=0;\n"
        "Park_SBML.parameters(1).fixed=0;\n"
        'Park_SBML.parameters(1).reaction="global";\n'
        ".....\n\n",
        "ruleAss": "Park_SBML.nruleAss=1;"
        'Park_SBML.ruleAss(1).id="D";\n'
        'Park_SBML.ruleAss(1).val="Feed / V";\n'
        ".....\n\n",
        "events": "Park_SBML.nevents=1;"
        'Park_SBML.events(1).id="Controller";\n'
        'Park_SBML.events(1).trigger="geq(time, 0)";\n'
        'Park_SBML.events(1).delay="0";\n'
        'Park_SBML.events(1).var(1)="Feed";\n'
        "Park_SBML.events(1).var(1)="
        ";\n"
        ".....\n\n",
        "reaction": "Park_SBML.nreaction=4;\n"
        'Park_SBML.reaction(1).id="rPM";\n'
        'Park_SBML.reaction(1).rate="vPM * (PT - PM)";\n'
        "Park_SBML.reaction(1).rPM=\"['1', '0', '0', '0']\";\n"
        ".....\n\n",
        "raterules": "Park_SBML.nraterules=1;\n"
        'Park_SBML.raterules(1).id="V";\n'
        'Park_SBML.raterules(1).val="D * V";\n'
        "Park_SBML.raterules(1).min=0;\n"
        "Park_SBML.raterules(1).max=1.5;\n"
        "Park_SBML.raterules(1).isres=0;\n"
        ".....\n\n",
    }

    return examples_dict[category]


def get_current_number(parameter):
    return int(parameter.split("(")[1].split(")")[0])


def check_parameters(correct_parameters, parameters, category, count):
    if sorted(parameters) != sorted(correct_parameters):
        if (
            category == "reaction"
            and "id" in parameters
            and "rate" in parameters
            and len(parameters) == 3
        ):
            return True

        # Check if there are any missing parameters
        missing_parameters = [i for i in correct_parameters if i not in parameters]

        if len(missing_parameters) > 0:
            return f"Error in {category} {count + 1} parameters. Parameters are missing: {missing_parameters}\n"
        else:
            return f"Error in {category} {count + 1} parameters. The correct list of parameters is {parameters}\n"
    else:
        return True


def get_value_from_parameter(parameter, lines):
    for i in lines:
        if parameter in i:
            return get_parameter_value(i)


def get_vars_from_event(lines):
    var = [get_parameter_value(i) for i in lines if "var(" in i]
    return var


def add_to_hmod(category, lines, hmod):
    if category == "species":
        hmod.add_specie(
            get_value_from_parameter(".id=", lines),
            get_value_from_parameter(".compartment=", lines),
            get_value_from_parameter(".val=", lines),
            get_value_from_parameter(".fixed=", lines),
            get_value_from_parameter(".min=", lines),
            get_value_from_parameter(".max=", lines),
            get_value_from_parameter(".isres=", lines),
        )
    elif category == "compartments":
        hmod.add_compartment(
            get_value_from_parameter(".id=", lines),
            get_value_from_parameter(".val=", lines),
            get_value_from_parameter(".fixed=", lines),
            get_value_from_parameter(".min=", lines),
            get_value_from_parameter(".max=", lines),
            get_value_from_parameter(".isres=", lines),
        )
    elif category == "parameters":
        hmod.add_parameter(
            get_value_from_parameter(".id=", lines),
            get_value_from_parameter(".reaction=", lines),
            get_value_from_parameter(".val=", lines),
            get_value_from_parameter(".fixed=", lines),
        )
    elif category == "ruleAss":
        hmod.add_assignment_rule(
            get_value_from_parameter(".id=", lines),
            get_value_from_parameter(".val=", lines),
        )
    elif category == "events":
        var = get_vars_from_event(lines)
        hmod.add_event(
            get_value_from_parameter(".id=", lines),
            get_value_from_parameter(".trigger=", lines),
            get_value_from_parameter(".delay=", lines),
            var[0],
            var[1],
        )
    elif category == "reaction":
        hmod.add_reaction(
            get_value_from_parameter(".id=", lines),
            get_value_from_parameter(".rate=", lines),
            get_value_from_parameter(
                f".{get_value_from_parameter('.id=', lines)}=", lines
            ),
        )
    elif category == "raterules":
        hmod.add_rate_rule(
            get_value_from_parameter(".id=", lines),
            get_value_from_parameter(".val=", lines),
            get_value_from_parameter(".min=", lines),
            get_value_from_parameter(".max=", lines),
            get_value_from_parameter(".isres=", lines),
        )


def divide_lines(lines, number):
    return [lines[i : i + number] for i in range(0, len(lines), number)]


def validate_category(category, lines, hmod):
    """
    Checks if the catagory is well writen
    :param hmod:
    :param category: The category of the parameters, i.e. Species, Compartments
    :param lines: The lines in the file
    :return: Bool that represents the validity
    """

    # Get the category lines present in the file
    category_lines = get_category_lines(category, lines)

    # The number of element this category has
    try:
        category_number = int(category_lines[0].split("=")[-1].split(";")[0])
    except Exception:
        return (
            f"Error in n{category}. Please confirm that this line is correctly written."
        )

    if (
        len(category_lines) == 1
        and category != "events"
        and category != "raterules"
        and category != "ruleAss"
    ):
        # O nspecies tem de ser 0. so deve ser possivel para eventos e rate rules
        return f"Error in {category}. Cannot have 0 values."
    if category_number == 0 and (
        category == "events" or category == "raterules" and category == "ruleAss"
    ):
        return 0

    parameter_dict = {
        "species": ["id", "compartment", "val", "fixed", "min", "max", "isres"],
        "compartments": ["id", "val", "fixed", "min", "max", "isres"],
        "parameters": ["id", "reaction", "val", "fixed"],
        "ruleAss": ["id", "val"],
        "events": ["id", "var", "var", "trigger", "delay"],
        "reaction": ["id", "rate"],
        "raterules": ["id", "val", "min", "max", "isres"],
    }
    parameter_dict_number = {
        "species": 7,
        "compartments": 6,
        "parameters": 4,
        "ruleAss": 2,
        "events": 5,
        "reaction": 3,
        "raterules": 5,
    }

    # Get the parameters of the category
    parameters = parameter_dict[category]

    # Divide the category into smaller lists, that represent an element
    divided_species = divide_lines(category_lines[1:], parameter_dict_number[category])

    # If the number of elements is different than the one stated
    if len(divided_species) != category_number:
        return (
            f"Error in n{category} value. It does not match with the rest of the {category} values.\n"
            f"It should be displayed as following:\n"
            f"{get_category_examples(category)}"
        )

    for count, i in enumerate(divided_species):
        current_specie = list(i)

        # Check if the parameters are present for this category
        correct_parameters = check_parameters(
            parameters, get_parameters_from_list(current_specie), category, count
        )
        if correct_parameters != True:
            return correct_parameters

        # Check every line of this category to make sure it is well written
        for parameter in current_specie:
            if not match_patterns(category, parameter, hmod):
                if not (category == "reaction" and '="[' in parameter):
                    return f'Error in {category} {count + 1}. Please check line:\n"{parameter}"\n'

            # Check if the current number is correct
            if count + 1 != get_current_number(parameter):
                return f"Error in {category} {get_current_number(parameter)}. The number should be {count + 1}"
        add_to_hmod(category, current_specie, hmod)

    return 0


def validate_reactions_values(file_lines, hmod):
    values_lines = [i for i in file_lines if '="[' in i]

    for value in values_lines:
        expected_value = hmod.reactions[get_current_number(value) - 1].id
        if not match_reaction_pattern(value, expected_value):
            return f"Error in reactions {get_current_number(value)}. The id does not match the values line."

    return 0


def validate_hmod(file_lines):
    hmod = Hmod()

    categories = [
        "compartments",
        "species",
        "parameters",
        "ruleAss",
        "events",
        "reaction",
        "raterules",
    ]

    for category in categories:
        valid = validate_category(category, file_lines, hmod)
        if valid != 0:
            return valid

    valid = validate_reactions_values(file_lines, hmod)
    if valid != 0:
        return valid

    return hmod.is_valid()
